#ifndef SQLNCLI_H
#define SQLNCLI_H

#include <stddef.h>
#include <sql.h>
#include <sqltypes.h>

/*
 * extra SQLServer values
 */

#define SQL_COPT_SS_BASE						1200

#define SQL_COPT_SS_PERF_DATA_LOG				(SQL_COPT_SS_BASE+12)
#define SQL_COPT_SS_MARS_ENABLED				(SQL_COPT_SS_BASE+24)

#define SQL_MARS_ENABLED_NO						0L
#define SQL_MARS_ENABLED_YES					1L

#define SQL_TXN_SS_SNAPSHOT						0x00000020L

#define SQL_COPT_SS_PRESERVE_CURSORS            (SQL_COPT_SS_BASE+4) /* Preserve server cursors after SQLTransact */
#define SQL_COPT_SS_INTEGRATED_SECURITY         (SQL_COPT_SS_BASE+3) /* Force integrated security on login */

#define SQL_BCP_OFF                             0L           /*  BCP is not allowed on connection */
#define SQL_BCP_ON                              1L           /*  BCP is allowed on connection */
#define SQL_BCP_DEFAULT                         SQL_BCP_OFF

#define SQL_COPT_SS_BCP                         (SQL_COPT_SS_BASE+19) /* Allow BCP usage on connection */

#define SQL_COPT_SS_ENLIST_IN_XA                SQL_ATTR_ENLIST_IN_XA /* Enlist in a XA transaction */

#define SQL_COPT_SS_TXN_ISOLATION				(SQL_COPT_SS_BASE+27)
#define SQL_COPT_SS_FAILOVER_PARTNER			(SQL_COPT_SS_BASE+25)
#define SQL_COPT_SS_TRUST_SERVER_CERTIFICATE	(SQL_COPT_SS_BASE+28)
#define SQL_COPT_SS_ENCRYPT						(SQL_COPT_SS_BASE+23)

#define SQL_EN_OFF                              0
#define SQL_EN_ON                               1

#define SQL_COPT_SS_SERVER_SPN                  (SQL_COPT_SS_BASE+29) /* Server SPN */
#define SQL_COPT_SS_FAILOVER_PARTNER_SPN        (SQL_COPT_SS_BASE+30) /* Failover partner server SPN */
#define SQL_COPT_SS_INTEGRATED_AUTHENTICATION_METHOD    (SQL_COPT_SS_BASE+31) /* The integrated authentication method used for the connection*/
#define SQL_COPT_SS_MUTUALLY_AUTHENTICATED      (SQL_COPT_SS_BASE+32) /* Used to decide if the connection is mutually authenticated */

#define SQL_COPT_SS_BASE_EX                     1240

#define SQL_COPT_SS_CONNECTION_DEAD				(SQL_COPT_SS_BASE_EX+4)
#define SQL_COPT_SS_APPLICATION_INTENT		    (SQL_COPT_SS_BASE_EX+7)
#define SQL_COPT_SS_MULTISUBNET_FAILOVER		(SQL_COPT_SS_BASE_EX+8)

#define SQL_COPT_SS_CONNECT_RETRY_COUNT         (SQL_COPT_SS_BASE_EX+20)
#define SQL_COPT_SS_CONNECT_RETRY_INTERVAL      (SQL_COPT_SS_BASE_EX+21)

typedef struct AccessToken
{
    DWORD dataSize;
    BYTE data[1];
} ACCESSTOKEN;

#define SQL_COPT_SS_ACCESS_TOKEN                (SQL_COPT_SS_BASE_EX+15)

#define SQL_PRESERVE_CURSORS                    SQL_COPT_SS_PRESERVE_CURSORS

#define SQL_COPT_SS_TXN_ISOLATION				(SQL_COPT_SS_BASE+27)

#define SQL_SOPT_SS_BASE                        1225

#define SQL_SOPT_SS_CURSOR_OPTIONS              (SQL_SOPT_SS_BASE+5) /* Server cursor options */
#define SQL_SOPT_SS_DEFER_PREPARE               (SQL_SOPT_SS_BASE+7) /* Defer prepare until necessary */

/* Defines for use with SQL_COPT_SS_INTEGRATED_SECURITY - Pre-Connect Option only */
#define SQL_IS_OFF                          0L           /*  Integrated security isn't used */
#define SQL_IS_ON                           1L           /*  Integrated security is used */
#define SQL_IS_DEFAULT                      SQL_IS_OFF

/* Defines for use with SQL_COPT_SS_PRESERVE_CURSORS */
#define SQL_PC_OFF                          0L           /*  Cursors are closed on SQLTransact */
#define SQL_PC_ON                           1L           /*  Cursors remain open on SQLTransact */
#define SQL_PC_DEFAULT                      SQL_PC_OFF

/* SQL_SOPT_SS_DEFER_PREPARE */
#define SQL_DP_OFF                          0L
#define SQL_DP_ON                           1L

/* Defines for use with SQL_SOPT_SS_CURSOR_OPTIONS */
#define SQL_CO_OFF                          0L           /*  Clear all cursor options */
#define SQL_CO_FFO                          1L           /*  Fast-forward cursor will be used */
#define SQL_CO_AF                           2L           /*  Autofetch on cursor open */
#define SQL_CO_FFO_AF                       (SQL_CO_FFO|SQL_CO_AF)  /*  Fast-forward cursor with autofetch */
#define SQL_CO_FIREHOSE_AF                  4L           /*  Auto fetch on fire-hose cursors */
#define SQL_CO_DEFAULT                      SQL_CO_OFF

/* Defines for use with SQL_SOPT_SS_HIDDEN_COLUMNS */
#define SQL_HC_OFF                          0L           /*  FOR BROWSE columns are hidden */
#define SQL_HC_ON                           1L           /*  FOR BROWSE columns are exposed */
#define SQL_HC_DEFAULT                      SQL_HC_OFF

/* SQL_SOPT_SS_NAME_SCOPE */
#define SQL_SS_NAME_SCOPE_TABLE             0L
#define SQL_SS_NAME_SCOPE_TABLE_TYPE        1L
#define SQL_SS_NAME_SCOPE_EXTENDED          2L
#define SQL_SS_NAME_SCOPE_SPARSE_COLUMN_SET 3L
#define SQL_SS_NAME_SCOPE_DEFAULT           SQL_SS_NAME_SCOPE_TABLE

/* SQLSetStmtAttr SQL Server Native Client driver specific defines. */
/* Statement attributes */
#define SQL_SOPT_SS_BASE                            1225
#define SQL_SOPT_SS_PARAM_FOCUS                     (SQL_SOPT_SS_BASE+11) /* Direct subsequent calls to parameter related methods to set properties on constituent columns/parameters of container types */
#define SQL_SOPT_SS_NAME_SCOPE                      (SQL_SOPT_SS_BASE+12) /* Sets name scope for subsequent catalog function calls */
#define SQL_SOPT_SS_MAX_USED                        SQL_SOPT_SS_NAME_SCOPE


/* 
 * SQL_SS_LENGTH_UNLIMITED is used to describe the max length of
 * VARCHAR(max), VARBINARY(max), NVARCHAR(max), and XML columns
 */
#define SQL_SS_LENGTH_UNLIMITED             0

/*
 * SQL Server 2008 new datatypes
 */

#define SQL_SS_VARIANT                      (-150)
#define SQL_SS_UDT                          (-151)
#define SQL_SS_XML                          (-152)
#define SQL_SS_TABLE                        (-153)
#define SQL_SS_TIME2                        (-154)
#define SQL_SS_TIMESTAMPOFFSET              (-155)

/* SQLGetDiagField driver specific defines. */
/* Microsoft has -1150 thru -1199 reserved for Microsoft SQL Server Native Client driver usage. */
#define SQL_DIAG_SS_BASE                    (-1150)
#define SQL_DIAG_SS_MSGSTATE                (SQL_DIAG_SS_BASE)
#define SQL_DIAG_SS_SEVERITY                (SQL_DIAG_SS_BASE-1)
#define SQL_DIAG_SS_SRVNAME                 (SQL_DIAG_SS_BASE-2)
#define SQL_DIAG_SS_PROCNAME                (SQL_DIAG_SS_BASE-3)
#define SQL_DIAG_SS_LINE                    (SQL_DIAG_SS_BASE-4)

/* SQLColAttribute Easysoft specific */
#define SQL_CA_SS_BASE                              1200
#define SQL_CA_SS_COLUMN_HIDDEN             (SQL_CA_SS_BASE+11)     /* column is hidden part of key */
#define SQL_CA_SS_COLUMN_KEY                (SQL_CA_SS_BASE+12)     /* column is part of primary key */
#define SQL_CA_SS_COLUMN_LCID               (SQL_CA_SS_BASE+37)     /* LCID of column */
#define SQL_CA_SS_COLUMN_SID                (SQL_CA_SS_BASE+38)     /* Sort Id of column */

#define SQL_COPT_SS_USE_PROC_FOR_PREP       (SQL_COPT_SS_BASE+2)  /* Use create proc for SQLPrepare */
#define SQL_COPT_SS_QUOTED_IDENT            (SQL_COPT_SS_BASE+17) /* Enable/Disable Quoted Identifiers */
#define SQL_COPT_SS_ANSI_NPW                (SQL_COPT_SS_BASE+18) /* Enable/Disable ANSI NULL, Padding and Warnings */

/* Query notification options */
#define SQL_SOPT_SS_QUERYNOTIFICATION_TIMEOUT   (SQL_SOPT_SS_BASE+8) 
#define SQL_SOPT_SS_QUERYNOTIFICATION_MSGTEXT   (SQL_SOPT_SS_BASE+9)
#define SQL_SOPT_SS_QUERYNOTIFICATION_OPTIONS   (SQL_SOPT_SS_BASE+10)

/* Severity codes for SQL_DIAG_SS_SEVERITY */
#define EX_ANY          0
#define EX_INFO         10
#define EX_MAXISEVERITY EX_INFO
#define EX_MISSING      11
#define EX_TYPE         12
#define EX_DEADLOCK     13
#define EX_PERMIT       14
#define EX_SYNTAX       15
#define EX_USER         16
#define EX_RESOURCE     17
#define EX_INTOK        18
#define MAXUSEVERITY    EX_INTOK
#define EX_LIMIT        19
#define EX_CMDFATAL     20
#define MINFATALERR     EX_CMDFATAL
#define EX_DBFATAL      21
#define EX_TABCORRUPT   22
#define EX_DBCORRUPT    23
#define EX_HARDWARE     24
#define EX_CONTROL      25


/* New Date Time Structures */
/* New Structure for TIME2 */
/* EASYSOFT ADDITION ============================================================================== */
/* There is a bit of a gotcha here, the normal packing would make this 12 bytes in length           */
/* but MS set the allignment to 2 bytes, so it becomes 10 bytes. If we forced this to 2 byte        */
/* in the driver, then there would need to be a platform specific packing pragma, so for the        */
/* moment, allow this to end up being size 12, this may change before the end of the beta depending */
/* on feedback we get                                                                               */
/* END EASYSOFT ADDITION ========================================================================== */
typedef struct tagSS_TIME2_STRUCT
    {
    SQLUSMALLINT hour;
    SQLUSMALLINT minute;
    SQLUSMALLINT second;
    SQLUINTEGER fraction;
    } 	SQL_SS_TIME2_STRUCT;

/* New Structure for TIMESTAMPOFFSET */
typedef struct tagSS_TIMESTAMPOFFSET_STRUCT
    {
    SQLSMALLINT year;
    SQLUSMALLINT month;
    SQLUSMALLINT day;
    SQLUSMALLINT hour;
    SQLUSMALLINT minute;
    SQLUSMALLINT second;
    SQLUINTEGER fraction;
    SQLSMALLINT timezone_hour;
    SQLSMALLINT timezone_minute;
    } 	SQL_SS_TIMESTAMPOFFSET_STRUCT;

#define SQL_SOPT_SS_HIDDEN_COLUMNS    (SQL_SOPT_SS_BASE+2) /* Expose FOR BROWSE hidden columns */
#define SQL_SOPT_SS_COLUMN_ENCRYPTION (SQL_SOPT_SS_BASE+13) /* Sets the column encryption mode */

/* Defines for use with SQL_SOPT_SS_COLUMN_ENCRYPTION */
#define SQL_CE_DISABLED                     0L           /*  Disabled */
#define SQL_CE_RESULTSETONLY                1L           /*  Decryption Only (resultsets and return values) */
#define SQL_CE_ENABLED                      3L           /*  Enabled (both encryption and decryption) */

/* force column encryption */
#define SQL_CA_SS_FORCE_ENCRYPT     (SQL_CA_SS_BASE+36) /*  indicate mandatory encryption for this parameter */

#define SQL_COPT_SS_CEKEYSTOREPROVIDER  (SQL_COPT_SS_BASE_EX+11) /* Load a keystore provider or read the list of loaded keystore providers */
#define SQL_COPT_SS_CEKEYSTOREDATA      (SQL_COPT_SS_BASE_EX+12) /* Communicate with loaded keystore providers */

/* Keystore Provider interface definition */
typedef struct CEKeystoreContext
{
    void *envCtx;
    void *dbcCtx;
    void *stmtCtx;
} CEKEYSTORECONTEXT;

typedef void errFunc(CEKEYSTORECONTEXT *ctx, const WCHAR *msg, ...);

typedef struct CEKeystoreProvider
{
    WCHAR *Name;
    int (*Init)(CEKEYSTORECONTEXT *ctx, errFunc *onError);
    int (*Read)(CEKEYSTORECONTEXT *ctx, errFunc *onError, void *data, unsigned int *len);
    int (*Write)(CEKEYSTORECONTEXT *ctx, errFunc *onError, void *data, unsigned int len);
    int (*DecryptCEK)(
        CEKEYSTORECONTEXT *ctx,
        errFunc *onError,
        const wchar_t *keyPath,
        const wchar_t *alg,
        unsigned char *ecek,
        unsigned short ecekLen,
        unsigned char **cekOut,
        unsigned short *cekLen);
    int(*EncryptCEK)(
        CEKEYSTORECONTEXT *ctx,
        errFunc *onError,
        const wchar_t *keyPath,
        const wchar_t *alg,
        unsigned char *cek,
        unsigned short cekLen,
        unsigned char **ecekOut,
        unsigned short *ecekLen);
    void (*Free)(void);
} CEKEYSTOREPROVIDER;

/* Communication between the driver and application via the CEKeystoreData structure */
typedef struct CEKeystoreData
{
    WCHAR *name;
    unsigned int dataSize;
    char data[1];
} CEKEYSTOREDATA;

/*
* BCP Definitions
*/

/* Error codes */
#define SUCCEED                 1
#define FAIL                    0
#define SUCCEED_ABORT           2
#define SUCCEED_ASYNC           3

/* Transfer directions */
#define DB_IN                   1   /* Transfer from client to server */
#define DB_OUT                  2   /* Transfer from server to client */

/* bcp_control option */
#define BCPMAXERRS              1   /* Sets max errors allowed */
#define BCPFIRST                2   /* Sets first row to be copied out */
#define BCPLAST                 3   /* Sets number of rows to be copied out */
#define BCPBATCH                4   /* Sets input batch size */
#define BCPKEEPNULLS            5   /* Sets to insert NULLs for empty input values */
#define BCPABORT                6   /* Sets to have bcpexec return SUCCEED_ABORT */
#define BCPODBC                 7   /* Sets ODBC canonical character output */
#define BCPKEEPIDENTITY         8   /* Sets IDENTITY_INSERT on */
#define BCPHINTSA               10  /* Sets server BCP hints (ANSI string) */
#define BCPHINTSW               11  /* Sets server BCP hints (UNICODE string) */
#define BCPFILECP               12  /* Sets clients code page for the file */
#define BCPUNICODEFILE          13  /* Sets that the file contains unicode header */
#define BCPTEXTFILE             14  /* Sets BCP mode to expect a text file and to detect Unicode or ANSI automatically */
#define BCPFILEFMT              15  /* Sets file format version */
#define BCPFMTXML               16  /* Sets the format file type to xml */
#define BCPFIRSTEX              17  /* Starting Row for BCP operation (64 bit) */
#define BCPLASTEX               18  /* Ending Row for BCP operation (64 bit) */
#define BCPROWCOUNT             19  /* Total Number of Rows Copied (64 bit) */
#define BCPDELAYREADFMT         20  /* Delay reading format file until bcp_exec */
/* BCPFILECP values
* Any valid code page that is installed on the client can be passed plus:
*/
#define BCPFILECP_RAW           (-1) /* Data in file is in Server code page (no conversion) */
/* bcp_collen definition */
#define SQL_VARLEN_DATA (-10)   /* Use default length for column */
/* BCP column format properties */
#define BCP_FMT_TYPE            0x01
#define BCP_FMT_INDICATOR_LEN   0x02
#define BCP_FMT_DATA_LEN        0x03
#define BCP_FMT_TERMINATOR      0x04
#define BCP_FMT_SERVER_COL      0x05
#define BCP_FMT_COLLATION       0x06
#define BCP_FMT_COLLATION_ID    0x07
/* bcp_setbulkmode properties */
#define BCP_OUT_CHARACTER_MODE      0x01
#define BCP_OUT_WIDE_CHARACTER_MODE 0x02
#define BCP_OUT_NATIVE_TEXT_MODE    0x03
#define BCP_OUT_NATIVE_MODE         0x04

#ifndef INT
typedef int     INT;
typedef INT     DBINT;
typedef DBINT * LPDBINT;
typedef unsigned char DBBOOL;

#ifndef _LPCBYTE_DEFINED
#define _LPCBYTE_DEFINED
typedef BYTE const* LPCBYTE;
#endif /* _LPCBYTE_DEFINED */
#endif /* INT */

/*
 * Data Types
 */

#ifndef SQLNCLI_NO_BCP
/* Define the symbol SQLNCLI_NO_BCP if you are not using BCP in your application */
/* and you want to exclude the BCP-related definitions in this header file. */

/* SQL Server Data Type defines. */
/* New types for SQL 6.0 and later servers */
#define SQLTEXT                             0x23
#define SQLVARBINARY                        0x25
#define SQLINTN                             0x26
#define SQLVARCHAR                          0x27
#define SQLBINARY                           0x2d
#define SQLIMAGE                            0x22
#define SQLCHARACTER                        0x2f
#define SQLINT1                             0x30
#define SQLBIT                              0x32
#define SQLINT2                             0x34
#define SQLINT4                             0x38
#define SQLMONEY                            0x3c
#define SQLDATETIME                         0x3d
#define SQLFLT8                             0x3e
#define SQLFLTN                             0x6d
#define SQLMONEYN                           0x6e
#define SQLDATETIMN                         0x6f
#define SQLFLT4                             0x3b
#define SQLMONEY4                           0x7a
#define SQLDATETIM4                         0x3a
/* New types for SQL 6.0 and later servers */
#define SQLDECIMAL                          0x6a
#define SQLNUMERIC                          0x6c
/* New types for SQL 7.0 and later servers */
#define SQLUNIQUEID                         0x24
#define SQLBIGCHAR                          0xaf
#define SQLBIGVARCHAR                       0xa7
#define SQLBIGBINARY                        0xad
#define SQLBIGVARBINARY                     0xa5
#define SQLBITN                             0x68
#define SQLNCHAR                            0xef
#define SQLNVARCHAR                         0xe7
#define SQLNTEXT                            0x63
/* New types for SQL 2000 and later servers */
#define SQLINT8                             0x7f
#define SQLVARIANT                          0x62
/* New types for SQL 2005 and later servers */
#define SQLUDT                              0xf0
#define SQLXML                              0xf1
/* New types for SQL 2008 and later servers */
#define SQLTABLE                            0xf3
#define SQLDATEN                            0x28
#define SQLTIMEN                            0x29
#define SQLDATETIME2N                       0x2a
#define SQLDATETIMEOFFSETN                  0x2b
/* Define old names */
#define SQLDECIMALN                         0x6a
#define SQLNUMERICN                         0x6c
#endif /* SQLNCLI_NO_BCP */


/*
* BCP functions
*/

DBINT SQL_API bcp_batch(HDBC);
RETCODE SQL_API bcp_bind(HDBC, LPCBYTE, INT, DBINT, LPCBYTE, INT, INT, INT);
RETCODE SQL_API bcp_colfmt(HDBC, INT, BYTE, INT, DBINT, LPCBYTE, INT, INT);
RETCODE SQL_API bcp_collen(HDBC, DBINT, INT);
RETCODE SQL_API bcp_colptr(HDBC, LPCBYTE, INT);
RETCODE SQL_API bcp_columns(HDBC, INT);
RETCODE SQL_API bcp_control(HDBC, INT, void *);
DBINT SQL_API bcp_done(HDBC);
RETCODE SQL_API bcp_exec(HDBC, LPDBINT);
RETCODE SQL_API bcp_getcolfmt(HDBC, INT, INT, void *, INT, INT *);
RETCODE SQL_API bcp_moretext(HDBC, DBINT, LPCBYTE);
RETCODE SQL_API bcp_sendrow(HDBC);
RETCODE SQL_API bcp_setbulkmode(HDBC, INT, void*, INT cbField, void *, INT cbRow);
RETCODE SQL_API bcp_setcolfmt(HDBC, INT, INT, void *, INT);

#ifdef UNICODE
#define bcp_init        bcp_initW
#define bcp_readfmt     bcp_readfmtW
#define bcp_writefmt    bcp_writefmtW
#define bcp_gettypename bcp_gettypenameW
#define dbprtype        dbprtypeW
#define BCPHINTS        BCPHINTSW
#else
#define bcp_init        bcp_initA
#define bcp_readfmt     bcp_readfmtA
#define bcp_writefmt    bcp_writefmtA
#define bcp_gettypename bcp_gettypenameA
#define dbprtype        dbprtypeA
#define BCPHINTS        BCPHINTSA
#endif /* UNICODE */

/* Narrow and wide character names functions */
CHAR* SQL_API bcp_gettypenameA(INT, DBBOOL);
WCHAR* SQL_API bcp_gettypenameW(INT, DBBOOL);
RETCODE SQL_API bcp_initA(HDBC, LPCSTR, LPCSTR, LPCSTR, INT);
RETCODE SQL_API bcp_initW(HDBC, LPCWSTR, LPCWSTR, LPCWSTR, INT);
RETCODE SQL_API bcp_readfmtA(HDBC, LPCSTR);
RETCODE SQL_API bcp_readfmtW(HDBC, LPCWSTR);
RETCODE SQL_API bcp_writefmtA(HDBC, LPCSTR);
RETCODE SQL_API bcp_writefmtW(HDBC, LPCWSTR);
CHAR* SQL_API dbprtypeA(INT);
WCHAR* SQL_API dbprtypeW(INT);

#endif
